@@ -1,7 +1,6 @@
 import argparse
 import asyncio
 import datetime
-import logging
 import os
 import pickle
 import random
@@ -9,31 +8,36 @@ import sys
 import time
 from threading import Thread
 from typing import Dict
+import numbers
 
 import numpy as np
-
 from stadle.lib.entity.model import BaseModel
 from stadle.lib.env.handler import EnvironmentHandler
-from stadle.lib.util import client_arg_parser
-from stadle.lib.util.debug import debug_with_stdout
-from stadle.lib.util.helpers import read_config, get_id, \
-    save_model_file, load_model_file, read_state, write_state, is_sgm_received, \
-    get_ip, compatible_data_dict_read, \
-    compatible_performance_dict_read, generate_performance_dict
-from stadle.lib.util.messengers import generate_lmodel_update_message, generate_participate_message, \
-    generate_initialize_model_message
+from stadle.lib.logging.logger import Logger
+from stadle.lib.util.bm_upload import get_base_model
+from stadle.lib.util.communication_handler import CommunicationHandler
+from stadle.lib.util.helpers import (compatible_data_dict_read,
+                                     compatible_performance_dict_read,
+                                     generate_performance_dict, get_id, get_ip,
+                                     is_sgm_received, load_model_file,
+                                     read_config, read_state, save_model_file,
+                                     write_state)
+from stadle.lib.util.messengers import (generate_initialize_model_message,
+                                        generate_lmodel_update_message,
+                                        generate_participate_message)
 from stadle.lib.util.model_conversion import extract_weights_dict
 # Client states
-from stadle.lib.util.states import ClientState, AgentMsgType, SGMDistributionMsgLocation, AggMsgType, \
-    BaseModelConvFormat # this import is ambiguos
-from stadle.lib.util.communication_handler import CommunicationHandler
+from stadle.lib.util.states import (AgentMsgType, AggMsgType, ClientState,
+                                    SGMDistributionMsgLocation)
 
-if (EnvironmentHandler.pt_installed_flag()):
+from stadle.lib.util.exceptions import (MaxAgentsReachedError,
+                                        MaxRoundReachedError,
+                                        MaxSizeReachedError,
+                                        AggRejectionError)
+
+if EnvironmentHandler.pt_installed_flag():
     from torch import nn
 
-logging.basicConfig(level=logging.INFO)
-
-
 class Client:
     """
     Client class instance provides the communication interface
@@ -124,7 +128,9 @@ class Client:
                  agent_running: bool = True,
                  max_round: int = -1):
 
-        logging.debug(f"Agent initialized at {time.time()}")
+        self.logger = Logger(self.__class__.__name__.__str__()).logger
+
+        self.logger.debug(f"Agent initialized at {time.time()}")
 
         # Waiting 1 sec, optimize this sleep time if needed
         wait_time = 1 * random.random()
@@ -138,13 +144,13 @@ class Client:
         # Read config
         if config_file is None:
             config_file = self._env_handler.agent_config_path
-            logging.warning(
+            self.logger.warning(
                 "NOTICE: No config file path was specified - using default agent config file")
-            logging.warning(
+            self.logger.warning(
                 "To use a specific config file, specify the path using the CLI argument --config_file when running the agent")
 
         self.config = read_config(config_file)
-        logging.debug(f'Config loaded: {self.config}')
+        self.logger.debug(f'Config loaded: {self.config}')
 
         default_config = read_config(self._env_handler.agent_config_path)
         default_keys = {k for k in default_config.keys() if k not in self.config.keys()}
@@ -152,15 +158,17 @@ class Client:
         for k in default_keys:
             self.config[k] = default_config[k]
 
-        if (default_keys):
-            logging.info('The following configuration parameters were not specified in the config file')
-            logging.info('Placeholder values will be taken from default config - these will be replaced by constructor arguments')
+        if default_keys:
+            self.logger.info('The following configuration parameters were not specified in the config file')
+            self.logger.info('Placeholder values will be taken from default config - these will be replaced by constructor arguments')
             for k in default_keys:
-                logging.info(f'\t{k} - setting to default value of {default_config[k]}')
+                self.logger.info(f'\t{k} - setting to default value of {default_config[k]}')
 
         # Check command line argvs
+        parse_flag = lambda s: (str(s).lower() == 'true') 
+
         if simulation_flag is None:
-            self.simulation_flag = bool(self.config['simulation'])
+            self.simulation_flag = parse_flag(self.config['simulation'])
         else:
             self.simulation_flag = simulation_flag
             self.config['simulation'] = simulation_flag
@@ -171,7 +179,7 @@ class Client:
             self.comm_protocol = comm_protocol
             self.config['comm_protocol'] = comm_protocol
 
-        self.comm_handler = CommunicationHandler(self.comm_protocol)
+        self.comm_handler = CommunicationHandler(self.comm_protocol, verbose=self.simulation_flag)
 
         # Comm. info to join the FL platform
         if aggregator_ip_address == None:
@@ -214,13 +222,31 @@ class Client:
             self.model_path = model_path + f'/{self.agent_name}'
             self.config['model_path'] = model_path + f'/{self.agent_name}'
 
-        debug_with_stdout(self.__class__.__name__.__str__(), "init", self.config)
 
-        if ('check_bm_type' in self.config):
-            self.check_bm_type = self.config['check_bm_type']
+        if 'check_bm_type' in self.config:
+            self.check_bm_type = parse_flag(self.config['check_bm_type'])
         else:
             self.check_bm_type = True
 
+        self.base_model = None
+
+        if 'base_model' in self.config:
+            bm_info = self.config['base_model']
+
+            if ('client_preload' not in bm_info or str(bm_info['client_preload']).lower() != 'false'):
+                bm_name = bm_info['model_name']
+                bm_fn = bm_info['model_fn']
+                bm_fn_src = bm_info['model_fn_src']
+                bm_format = bm_info['model_format']
+                if ('model_fn_args' in bm_info):
+                    bm_func_args = bm_info['model_fn_args']
+                else:
+                    bm_func_args = None
+                self.base_model = get_base_model(
+                    bm_name, bm_fn_src, bm_fn, bm_func_args, bm_format
+                )
+                self.logger.info(f'Loaded base model {bm_name} from config')
+
         # Get agent's IP address
         self.agent_ip = get_ip()
 
@@ -243,7 +269,7 @@ class Client:
         self.waiting_flag = ClientState.waiting_sgm
 
         # Initialization
-        self.init_weights_flag = bool(self.config['init_weights_flag'])
+        self.init_weights_flag = parse_flag(self.config['init_weights_flag'])
 
         self.version = 1.0
 
@@ -259,16 +285,15 @@ class Client:
 
         self.max_iterations = max_round if max_round > 0 else 10000
 
-        self.base_model = None
-
         self.num_samples = 1
 
         self.register_agent()
 
         if self.agent_running == True:
+            self.logger.info(f'--- start model exchange ---')
             self.start_model_exchange_routine()
 
-        logging.debug(f'--- Your IP is {self.agent_ip} ---')
+        self.logger.info(f'--- Initialization is done, your IP is {self.agent_ip} ---')
 
     async def participate(self):
         """
@@ -296,7 +321,7 @@ class Client:
                 self.round = resp[1]
                 self.exch_port = resp[2]
                 self.msend_port = resp[3]
-                logging.debug(f"--- Init Response: {resp[0]} ---")
+                self.logger.debug(f"--- Aggregator Response: {resp[0]} ---")
                 # State transition to waiting_sgm
                 self.tran_state(ClientState.waiting_sgm)
 
@@ -307,24 +332,40 @@ class Client:
                 self.round = resp[1]
                 self.exch_port = resp[2]
                 self.msend_port = resp[3]
-                logging.debug(f"--- Init Response: {resp[0]} ---")
+                self.logger.debug(f"--- Aggregator Response: {resp[0]} ---")
                 self.save_model_from_message(resp[4])
 
+            elif resp[0] == AggMsgType.max_agents_reached:
+                self.isRegistered = False
+                self.logger.error(f"--- Aggregator Response: {resp[1]}, system exits. ---")
+                raise MaxAgentsReachedError("The maximum number of agents are already connected to the aggregator - please increase the max_agents project limit, or stop some agents with timeout enabled to connect this agent to the aggregator.")
+
+            elif resp[0] == AggMsgType.max_round_reached:
+                self.isRegistered = False
+                self.logger.error(f"--- Aggregator Response: {resp[1]}, system exits. ---")
+                raise MaxRoundReachedError("The aggregator has already reached the maximum round - please increase the max_rounds project limit to continue this FL process")
+
             elif resp[0] == AggMsgType.registration_failed:
                 self.isRegistered = False
-                logging.error(f"--- Init Response: {resp[1]}, system exits. ---")
-                sys.exit()
+                self.logger.error(f"--- Aggregator Response: {resp[1]}, system exits. ---")
+                raise AggRejectionError(f"Registration failed with response {resp[1]}")
+
+            else:
+                self.logger.error(f"--- Aggregator Response: {resp[1]}, Unknown ---")
+                raise AggRejectionError(f"Registration failed with response {resp[1]}")
 
         except Exception as e:
-            logging.error(f"--- Cannot connect to the aggregator \n\t{e}---")
-            sys.exit()
+            self.logger.error(f"--- Cannot connect to the aggregator: {e} ---")
+            raise
 
     async def send_initial_models(self):
         """
         Send a message to initialize the first model
         :return:
         """
-        logging.debug(f'--- sending initial model. current round: {self.round} ---')
+        self.logger.info("Uploading base model to aggregator...")
+
+        self.logger.debug(f'--- sending initial model. current round: {self.round} ---')
 
         # Read the local models to tell the structure to the aggregator
         # (not necessarily trained)
@@ -332,24 +373,34 @@ class Client:
             data_dict, performance_dict = load_model_file(
                 self.model_path, self.lmfile)
         except:
-            logging.error("Cannot load local models!")
+            self.logger.error("Cannot load local models!")
             return
 
         rand_id, gene_time, models, model_id, self.version, base_model = \
             compatible_data_dict_read(data_dict)
-        logging.debug(models)
-
-        performance_id, performance_dict = compatible_performance_dict_read(performance_dict)
-        logging.debug(models)
+        self.logger.debug(models)
 
         wsaddr = f'{self.wsprefix}{self.reg_port}'
         try:
             # Create an initialize model message
             msg = generate_initialize_model_message(self, model_id, models, gene_time, performance_dict,
                                                     base_model)
-            await self.comm_handler.send_msg(msg, f'{self.aggr_ip}:{self.reg_port}', resp_required=False)
-        except:
-            logging.error("--- cannot connect to the aggregator ---")
+            resp = await self.comm_handler.send_msg(msg, f'{self.aggr_ip}:{self.reg_port}', resp_required=True)
+
+            if resp[0] == AggMsgType.max_size_reached:
+                self.isRegistered = False
+                self.logger.error(f"--- Aggregator Response: {resp[1]}, system exits. ---")
+                raise MaxSizeReachedError(resp[1])
+            
+            elif resp[0] == AggMsgType.not_opt:
+                self.isRegistered = False
+                self.logger.error(f"--- Aggregator Response: {resp[1]} ---")
+                raise AggRejectionError(resp[1])
+            
+            self.logger.info("Base model uploaded to aggregator")
+        except Exception as e:
+            self.logger.error(f"--- Cannot connect to the aggregator: {e} ---")
+            raise
 
         # TODO Receive models in a resp message as one option
         # State transition to waiting_sgm
@@ -362,19 +413,20 @@ class Client:
         :return:
         """
         if self.is_finite == True:
-            logging.debug(f'--- this is a finite model exchange routine with max iterations {self.max_iterations} ---')
+            self.logger.debug(f'--- this is a finite model exchange routine with max iterations {self.max_iterations} ---')
         else:
-            logging.debug(f'--- this is an infinite model exchange routine ---')
+            self.logger.debug(f'--- this is an infinite model exchange routine ---')
 
         loop_count = 0
-        while True:
+        while self.exch_active:
+
             # Number of iterations if clearly specified by users
             if self.is_finite == True and loop_count == self.max_iterations:
                 self.agent_running = False
 
             # stop runnign if the flag is False
             if not self.agent_running:
-                logging.debug(f"--- system stops.. ----")
+                self.logger.debug(f"--- system stops.. ----")
                 return
 
             # Periodically check the state
@@ -383,7 +435,7 @@ class Client:
             # ready to send models
             if state == ClientState.sending:
 
-                logging.debug('--- Local model is ready to be sent to the aggregator ---')
+                self.logger.debug('--- Local model is ready to be sent to the aggregator ---')
 
                 # Read the models from the local file
                 data_dict, performance_dict = load_model_file(
@@ -392,24 +444,43 @@ class Client:
                 _, _, models, model_id, _, _ = compatible_data_dict_read(
                     data_dict)
 
-                performance_id, performance_dict = compatible_performance_dict_read(performance_dict)
                 upd_msg = generate_lmodel_update_message(
                     self.id, model_id, models, performance_dict, self.num_samples)
 
                 try:
-                    await self.comm_handler.send_msg(upd_msg, f'{self.aggr_ip}:{self.msend_port}', resp_required=False)
-                    logging.debug('--- Local Models Sent ---')
+                    resp = await self.comm_handler.send_msg(upd_msg, f'{self.aggr_ip}:{self.msend_port}', resp_required=True)
+                    
+                    if resp[0] == AggMsgType.max_agents_reached:
+                        self.logger.error(f"--- Aggregator Response: {resp[1]}---")
+                        raise MaxAgentsReachedError(resp[1])
 
-                    # State transition to waiting_sgm
-                    self.tran_state(ClientState.waiting_sgm)
-                    logging.debug('--- state transition to waiting_sgm ---')
+                    elif resp[0] == AggMsgType.max_round_reached:
+                        self.logger.error(f"--- Aggregator Response: {resp[1]} ---")
+                        raise MaxRoundReachedError(resp[1])
+
+                    elif resp[0] == AggMsgType.not_opt:
+                        self.logger.error(f"--- Aggregator Response: {resp[1]} ---")
+                        raise AggRejectionError(resp[1])
+                    
+                    elif resp[0] == AggMsgType.max_size_reached:
+                        self.isRegistered = False
+                        self.logger.error(f"--- Aggregator Response: {resp[1]}, system exits. ---")
+                        raise MaxSizeReachedError(resp[1])
+
+                    else:
+                        self.logger.debug('--- Local Models Sent ---')
+
+                        # State transition to waiting_sgm
+                        self.tran_state(ClientState.waiting_sgm)
+                        self.logger.debug('--- state transition to waiting_sgm ---')
                 except Exception as e:
-                    logging.error(f"--- connection lost with the aggregator ---\n{e}")
+                    self.logger.error(f"--- Error during local model upload: {e} ---")
+                    raise
 
             # Polling for sg model update
             elif state == ClientState.waiting_sgm or state == ClientState.training:
 
-                logging.debug(f'--- polling to see if there is any update ---')
+                self.logger.debug(f'--- polling to see if there is any update ---')
 
                 msg = list()
                 msg.append(AgentMsgType.waiting_sgm)  # 0
@@ -421,12 +492,26 @@ class Client:
                     resp = await self.comm_handler.send_msg(msg, f'{self.aggr_ip}:{self.msend_port}')
 
                     if resp[0] == AggMsgType.update:
-                        logging.debug(f'--- SG Model Received ---')
+                        self.logger.debug(f'--- SG Model Received ---')
                         self.save_model_from_message(resp)
+
+                    elif resp[0] == AggMsgType.max_agents_reached:
+                        self.logger.error(f"--- Aggregator Response: {resp[1]}---")
+                        raise MaxAgentsReachedError(resp[1])
+
+                    elif resp[0] == AggMsgType.max_round_reached:
+                        self.logger.error(f"--- Aggregator Response: {resp[1]} ---")
+                        raise MaxRoundReachedError(resp[1])
+
+                    elif resp[0] == AggMsgType.not_opt:
+                        self.logger.error(f"--- Aggregator Response: {resp[1]} ---")
+                        raise AggRejectionError(resp[1])
+
                     else:
-                        logging.debug(f'--- SG Model is NOT ready ---')
+                        self.logger.debug(f'--- SG Model is NOT ready ---')
                 except Exception as e:
-                    logging.error(f"--- connection lost with the aggregator ---\n{e}")
+                    self.logger.error(f"--- Error during waiting_sgm poll: {e} ---")
+                    raise
 
             # just increment the loop count
             loop_count += 1
@@ -475,8 +560,8 @@ class Client:
         :return:
         """
         sgm_msg = pickle.loads(await websocket.recv())
-        logging.debug(f'--- SG Model Received ---')
-        logging.debug(f'Models: {sgm_msg}')
+        self.logger.debug(f'--- SG Model Received ---')
+        self.logger.debug(f'Models: {sgm_msg}')
         self.save_model_from_message(sgm_msg)
 
     # ---- APIs ---- #
@@ -519,7 +604,7 @@ class Client:
     def register_agent(self):
 
         if self.simulation_flag == True:
-            print(f"Default token [{self.token}] is used...")
+            self.logger.info(f"Default token [{self.token}] is used...")
 
         loop = asyncio.new_event_loop()
         asyncio.set_event_loop(loop)
@@ -527,9 +612,11 @@ class Client:
 
         # Redundant but just in case...
         if self.isRegistered == False:
-            logging.debug("Registration unsuccessful")
+            self.logger.error("Registration unsuccessful")
             sys.exit()
 
+        self.logger.info("Registration is successful")
+
     def initialize_models(self):
         loop = asyncio.new_event_loop()
         asyncio.set_event_loop(loop)
@@ -561,7 +648,7 @@ class Client:
             local_model_file_name = self.lmfile
 
         save_model_file(models, model_path, local_model_file_name, performance_dict)
-        logging.debug(f'--- Model Saved ---')
+        self.logger.debug(f'--- Model Saved ---')
 
     def send_model(self, models=None, performance_dict=dict(), model_path=None, local_model_file_name=None):
 
@@ -599,17 +686,19 @@ class Client:
 
     # API function for admin agent to perform base model init/update through aggregator
     def load_base_model(self, base_model: BaseModel):
-        if (base_model.type == None):
-            logging.debug(
+        if base_model.type is None:
+            self.logger.debug(
                 "Resolving model format from model object is currently not supported - please specify the model "
                 "format using \'BaseModelConvFormat.pytorch\' or \'BaseModelConvFormat.pytorch\'")
             raise ValueError("Ambiguous model conversion from unknown format")
 
-        if (base_model.name == None):
+        if base_model.name is None:
             gen_time = time.time()
             formatted_time = datetime.datetime.fromtimestamp(gen_time).strftime('%m-%d-%Y_%H:%M:%S')
             base_model.name = f'model_{formatted_time}'
 
+        self.logger.info(f"Preparing base model {base_model.name}...")
+
         base_model.prep_for_pickle()
 
         model_dict = {
@@ -618,6 +707,8 @@ class Client:
         }
         self.save_model(model_dict)
 
+        self.logger.info(f"Base model prepared for upload")
+
     def set_num_samples(self, num_samples):
         self.num_samples = num_samples
 
@@ -650,31 +741,29 @@ class BasicClient(Client):
                  agent_name: str = None,
                  model_path: str = None,
                  agent_running: bool = None,
-                 use_cl_args: bool = False
+                 cl_args = None
                  ):
-        if (use_cl_args):
-            args = client_arg_parser()
-
-            if args.config_path:
-                config_file = args.config_path
-            if args.simulation is not None:
-                simulation_flag = args.simulation
-            if args.comm_protocol:
-                comm_protocol = args.comm_protocol
-            if args.aggregator_ip:
-                aggregator_ip_address = args.aggregator_ip
-            if args.reg_port:
-                reg_port = args.reg_port
-            if args.exch_port:
-                exch_port = args.exch_port
-            if args.token:
-                token = args.token
-            if args.model_path:
-                model_path = args.model_path
-            if args.agent_name:
-                agent_name = args.agent_name
-            if args.agent_running:
-                agent_running = args.agent_running
+        if cl_args is not None:
+            if cl_args.config_path:
+                config_file = cl_args.config_path
+            if cl_args.simulation:
+                simulation_flag = cl_args.simulation
+            if cl_args.comm_protocol:
+                comm_protocol = cl_args.comm_protocol
+            if cl_args.aggregator_ip:
+                aggregator_ip_address = cl_args.aggregator_ip
+            if cl_args.reg_port:
+                reg_port = cl_args.reg_port
+            if cl_args.exch_port:
+                exch_port = cl_args.exch_port
+            if cl_args.token:
+                token = cl_args.token
+            if cl_args.model_path:
+                model_path = cl_args.model_path
+            if cl_args.agent_name:
+                agent_name = cl_args.agent_name
+            if cl_args.agent_running:
+                agent_running = cl_args.agent_running
 
         if agent_running is None:
             agent_running = True
@@ -685,9 +774,17 @@ class BasicClient(Client):
                          model_path=model_path, agent_running=agent_running)
 
     def disconnect(self):
+        time.sleep(5)
+        while (self.comm_handler.is_sending):
+            time.sleep(1)
+        self.logger.info('Disconnecting')
         self.stop_model_exchange_routine()
 
     def send_trained_model(self, trained_model, perf_values = {}):
+        for metric_key, metric_value in perf_values.items():
+            if (not isinstance(metric_value, numbers.Number)):
+                raise ValueError(f'Metric {metric_key} must be numeric type, not {type(metric_value)}')
+
         # Extract weights from trained local model
         trained_model_dict = extract_weights_dict(trained_model, self.get_base_model().type)
 
@@ -695,15 +792,13 @@ class BasicClient(Client):
 
         if client_state == ClientState.sg_ready:
             # Client received new SG weights before local training could complete
-            logging.debug(f'--- The training was too slow. A new SG model is available. ---')
+            self.logger.debug(f'--- The training was too slow. A new SG model is available. ---')
             return False
         else:
-            performance_dict = self.create_performance_data(**perf_values)
-
             # Save trained model weights and performance metrics, then send to aggregator
-            self.save_model(trained_model_dict, performance_dict)
+            self.save_model(trained_model_dict, perf_values)
             self.send_model()
-            logging.debug(f'--- Normal transition: The trained local models saved ---')
+            self.logger.debug(f'--- Normal transition: The trained local models saved ---')
 
             # Change client state to begin local model transmission
             self.tran_state(ClientState.sending)
@@ -739,31 +834,29 @@ class IntegratedClient(Client):
                  agent_name: str = None,
                  model_path: str = None,
                  agent_running: bool = None,
-                 use_cl_args: bool = False
+                 cl_args = None
                  ):
-        if (use_cl_args):
-            args = client_arg_parser()
-
-            if args.config_path:
-                config_file = args.config_path
-            if args.simulation is not None:
-                simulation_flag = args.simulation
-            if args.comm_protocol:
-                comm_protocol = args.comm_protocol
-            if args.aggregator_ip:
-                aggregator_ip_address = args.aggregator_ip
-            if args.reg_port:
-                reg_port = args.reg_port
-            if args.exch_port:
-                exch_port = args.exch_port
-            if args.token:
-                token = args.token
-            if args.model_path:
-                model_path = args.model_path
-            if args.agent_name:
-                agent_name = args.agent_name
-            if args.agent_running:
-                agent_running = args.agent_running
+        if cl_args is not None:
+            if cl_args.config_path:
+                config_file = cl_args.config_path
+            if cl_args.simulation:
+                simulation_flag = cl_args.simulation
+            if cl_args.comm_protocol:
+                comm_protocol = cl_args.comm_protocol
+            if cl_args.aggregator_ip:
+                aggregator_ip_address = cl_args.aggregator_ip
+            if cl_args.reg_port:
+                reg_port = cl_args.reg_port
+            if cl_args.exch_port:
+                exch_port = cl_args.exch_port
+            if cl_args.token:
+                token = cl_args.token
+            if cl_args.model_path:
+                model_path = cl_args.model_path
+            if cl_args.agent_name:
+                agent_name = cl_args.agent_name
+            if cl_args.agent_running:
+                agent_running = cl_args.agent_running
 
         if agent_running is None:
             agent_running = True
@@ -863,7 +956,7 @@ class IntegratedClient(Client):
                 # Get example performance metrics of SG model
                 acc, loss_test = self._validate(merged_model, self._validation_data,
                                                       **self._validation_kwargs)
-                logging.debug(f'--- SG Training accuracy: {acc} ---')
+                self.logger.debug(f'--- SG Training accuracy: {acc} ---')
 
             # Client ready to begin local training process
             elif client_state == ClientState.training:
@@ -873,26 +966,26 @@ class IntegratedClient(Client):
                 # Extract weights from trained local model
                 trained_model_dict = extract_weights_dict(trained_model, self.get_base_model().type)
                 self._training_count += 1
-                logging.debug(f'--- Training Done ---')
+                self.logger.debug(f'--- Training Done ---')
 
                 # Check if client state changed during local training
                 client_state = self.read_agent_state()
 
                 if client_state == ClientState.sg_ready:
                     # Client received new SG weights before local training could complete
-                    logging.debug(f'--- The training was too slow. A new set of SG modes are available. ---')
+                    self.logger.debug(f'--- The training was too slow. A new set of SG modes are available. ---')
                 else:
                     # Compute performance of local trained model on test dataset
                     acc, loss_test = self._validate(trained_model, self._validation_data,
                                                           **self._validation_kwargs)
                     # Create dict to store performance metrics for DB storage, GUI viewing
                     performance_dict = self.create_performance_data(False, acc, acc, loss_training, 0.0, loss_test)
-                    logging.debug(f'--- Training accuracy: {acc} ---')
+                    self.logger.debug(f'--- Training accuracy: {acc} ---')
 
                     # Save trained model weights and performance metrics, then send to aggregator
                     self.save_model(trained_model_dict, performance_dict)
                     self.send_model()
-                    logging.debug(f'--- Normal transition: The trained local models saved ---')
+                    self.logger.debug(f'--- Normal transition: The trained local models saved ---')
 
                     # Change client state to begin local model transmission
                     self.tran_state(ClientState.sending)
